from collections import defaultdict, Counter


class Policy:  # TODO: PolicyPredictor
    model = None
    observation_action_visit_count = defaultdict(lambda: Counter())
    observation_action_returns_sum = defaultdict(lambda: defaultdict(float))
    observation_action_value = defaultdict(lambda: defaultdict(float))

    def predict_action(
        self,
        observation,
        available_actions,
        available_action_probabilities=None,
        epsilon=None,  # TODO: use epsilon-greedy approach
    ):
        return max(  # TODO: randomly choose among duplicates
            range(len(available_actions)),
            key=lambda index: self.observation_action_value[observation][
                available_actions[index]
            ],
        )

    def learn(
        self, trajectories, gamma, alpha
    ):  # TODO: this impls every-visit MC, also show first-visit?
        for trajectory in trajectories:
            for i, (observation, action, reward) in enumerate(trajectory):
                sum_of_discounted_rewards = sum(
                    [
                        gamma ** (j - i) * trajectory[j][-1]
                        for j in range(i, len(trajectory))
                    ]
                )

                self.observation_action_visit_count[observation][action] += 1
                self.observation_action_returns_sum[observation][
                    action
                ] += sum_of_discounted_rewards
                self.observation_action_value[observation][action] = (
                    self.observation_action_returns_sum[observation][action]
                    / self.observation_action_visit_count[observation][action]
                )
